import itertools
import sys
from typing import (
    Any,
    AsyncIterator,
    Iterable,
    Iterator,
    Sequence,
    Tuple,
    TypeVar,
    no_type_check,
)

from langchain_core.documents import BaseDocumentTransformer, Document

from langchain_rag.document_transformers.lazy_document_transformer import (
    LazyDocumentTransformer,
)

BATCH_SIZE = 100

T = TypeVar("T")  # Only in python 3.12
if sys.version_info.major > 3 or sys.version_info.minor >= 12:
    from itertools import batched  # type: ignore[attr-defined]
else:

    def batched(iterable: Iterable[T], n: int) -> Iterator[Tuple[T, ...]]:
        if n < 1:
            raise ValueError("n must be at least one")
        it = iter(iterable)
        while batch := tuple(itertools.islice(it, n)):
            yield batch


async def async_batched(
    async_iterator: AsyncIterator[T], n: int
) -> AsyncIterator[AsyncIterator[T]]:
    if n < 1:
        raise ValueError("n must be at least one")

    async def _async_iterator(iterable: Iterable[Any]) -> AsyncIterator[Any]:
        for i in iterable:
            yield i

    batch = []
    i = 0
    async for item in async_iterator:
        i += 1
        batch.append(item)
        if i == n:
            yield _async_iterator(batch)  # type: ignore
            i = 0
            batch.clear()
    if batch:
        yield _async_iterator(batch)  # type: ignore


class DocumentTransformers(LazyDocumentTransformer):
    """
    Document transformer that uses a list of Transformers.
    Take each input document, and apply all transformations present in the
    `transformers` attribute.

    This is the basis for multiple transformations, using the plus operator.
    """

    def __init__(
        self,
        transformers: Sequence[BaseDocumentTransformer],
        batch_size: int = BATCH_SIZE,
    ):
        self.transformers = transformers
        self.batch_size = batch_size

    def __add__(
        self,
        other: LazyDocumentTransformer,
    ) -> "DocumentTransformers":
        """Compose this runnable with another object to create a RunnableSequence."""
        if isinstance(other, DocumentTransformers):
            return DocumentTransformers(
                transformers=list(other.transformers) + list(self.transformers),
            )
        else:
            return DocumentTransformers(transformers=list(self.transformers) + [other])

    @no_type_check  # Bug in Mypy
    def lazy_transform_documents(
        self, documents: Iterator[Document], **kwargs: Any
    ) -> Iterator[Document]:
        """Transform an iterator of documents with the list of transformations.

        Args:
            documents: A sequence of Documents to be transformed.

        Returns:
            An iterator oftransformed Documents.
        """
        # Can be refactored to use parallelism
        # Implementation when all transformers are NOT compatible with Runnable
        # It's not compatible with lazy strategy. Load all documents and apply
        # all transformations.
        for batch in batched(documents, self.batch_size):
            for t in self.transformers:
                for doc in t.transform_documents(documents=list(batch)):
                    yield doc

    @no_type_check  # Bug in Mypy
    async def _alazy_transform_documents(
        self, documents: AsyncIterator[Document], **kwargs: Any
    ) -> AsyncIterator[Document]:
        """Asynchronously transform an iterator of documents with a list
        of transformations.

        Args:
            documents: An iterator of Documents to be transformed.

        Returns:
            An iterator of transformed Documents.
        """
        # Implementation when all transformers are NOT compatible with Runnable
        # It's not compatible with lazy strategy. Load all documents and apply
        # all transformations.
        async for batch in async_batched(documents, self.batch_size):
            sync_batch = [doc async for doc in batch]
            for transformer in self.transformers:
                for doc in await transformer.atransform_documents(sync_batch):
                    yield doc



from itertools import chain
from typing import Any, AsyncIterator, Iterator, Sequence, cast

from langchain_core.documents import BaseDocumentTransformer, Document

from .document_transformers import BATCH_SIZE, async_batched
from .lazy_document_transformer import (
    LazyDocumentTransformer,
)


class DocumentTransformerPipeline(LazyDocumentTransformer):

    """List of document transformers that are chained together and run in sequence."""

    def __init__(
        self,
        transformers: Sequence[BaseDocumentTransformer],
        batch_size: int = BATCH_SIZE,
    ):
        self.transformers = transformers
        self.batch_size = batch_size

    def _lazy_transform_documents_with_transformer(
        self,
        documents: Iterator[Document],
        transformer: BaseDocumentTransformer,
        **kwargs: Any,
    ) -> Iterator[Document]:
        lazy_results = []
        for _document in documents:
            if hasattr(transformer, "lazy_transform_documents"):
                lazy_results.append(
                    transformer.lazy_transform_documents([_document], **kwargs)
                )
            else:
                lazy_results.append(
                    iter(transformer.transform_documents([_document], **kwargs))
                )
        return chain(*lazy_results)

    def lazy_transform_documents(
        self, documents: Iterator[Document], **kwargs: Any
    ) -> Iterator[Document]:
        if not self.transformers:
            return iter(documents)
        for _transformer in self.transformers:
            documents = self._lazy_transform_documents_with_transformer(
                documents, transformer=_transformer
            )
        return documents

    async def _alazy_transform_documents_with_transformer(
        self,
        documents: AsyncIterator[Document],
        transformer: BaseDocumentTransformer,
        **kwargs: Any,
    ) -> AsyncIterator[Document]:
        if hasattr(transformer, "alazy_transform_documents"):
            async for doc in transformer.alazy_transform_documents(documents, **kwargs):
                yield doc

        else:
            async for batch in async_batched(documents, self.batch_size):
                sync_batch = cast(Sequence[Document], [doc async for doc in batch])
                for doc in transformer.transform_documents(sync_batch, **kwargs):
                    yield doc

    async def _alazy_transform_documents(  # type: ignore
        self, documents: AsyncIterator[Document], **kwargs: Any
    ) -> AsyncIterator[Document]:
        for _transformer in self.transformers:
            async for doc in self._alazy_transform_documents_with_transformer(
                documents, transformer=_transformer
            ):
                yield doc



import asyncio
import threading
import time
from abc import abstractmethod
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterator,
    Iterator,
    Sequence,
    TypeVar,
    Union,
)

from langchain_core.documents import BaseDocumentTransformer, Document

"""
    We propose an alternative way of making transformers compatible with LCEL.
    The first keeps the current protocol (RunnableDocumentTransformer).
    The second takes advantage of this to propose
    a lazy approach to transformations (_RunnableGeneratorDocumentTransformer).
    It's better for the memory, pipeline, etc.

    Now, it's possible to create a pipeline of transformer like:
    Example:
    ..code - block:: python
    class UpperTransformer(_RunnableGeneratorDocumentTransformer):
        def lazy_transform_documents(
                self,
                documents: Iterator[Document],
                **kwargs: Any
        ) -> Iterator[Document]:
            ...

        async def alazy_transform_documents(
                self,
                documents: Union[AsyncIterator[Document],Iterator[Document]],
                **kwargs: Any
        ) -> AsyncIterator[Document]:
            ...

    runnable = (UpperTransformer() | LowerTransformer())
    result = list(runnable.invoke(documents))
"""

T = TypeVar("T")


async def to_async_iterator(iterator: Iterator[T]) -> AsyncIterator[T]:
    """Convert an iterable to an async iterator."""
    for item in iterator:
        yield item


_DONE = ""
_TIMEOUT = 1


def to_sync_iterator(async_iterable: AsyncIterator[T], maxsize: int = 0) -> Iterator[T]:
    def _run_coroutine(
        loop: asyncio.AbstractEventLoop,
        async_iterable: AsyncIterator[T],
        queue: asyncio.Queue,
    ) -> None:
        async def _consume_async_iterable(
            async_iterable: AsyncIterator[T], queue: asyncio.Queue
        ) -> None:
            async for x in async_iterable:
                await queue.put(x)

            await queue.put(_DONE)

        loop.run_until_complete(_consume_async_iterable(async_iterable, queue))

    queue: asyncio.Queue[T] = asyncio.Queue(maxsize=maxsize)
    loop = asyncio.new_event_loop()

    t = threading.Thread(target=_run_coroutine, args=(loop, async_iterable, queue))
    t.daemon = True
    t.start()

    while True:
        if not queue.empty():
            x = queue.get_nowait()

            if x is _DONE:
                break
            else:
                yield x
        else:
            time.sleep(_TIMEOUT)

    t.join()


Input = Union[AsyncIterator[Document], Iterator[Document], Sequence[Document]]
Output = Union[AsyncIterator[Document], Iterator[Document]]


class LazyDocumentTransformer(BaseDocumentTransformer):
    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        # Convert lazy to classical transformation
        return list(self.lazy_transform_documents(iter(documents), **kwargs))

    async def atransform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        return [
            doc
            async for doc in self.alazy_transform_documents(iter(documents), **kwargs)
        ]

    @staticmethod
    def _doc_to_async_iterator(
        documents: Sequence[Document] | AsyncIterator[Document] | Iterator[Document],
    ) -> AsyncIterator[Document]:
        if isinstance(documents, Sequence):
            async_documents = to_async_iterator(iter(documents))
        elif isinstance(documents, AsyncIterator):
            async_documents = documents
        elif isinstance(documents, Iterator):
            async_documents = to_async_iterator(documents)
        else:
            raise ValueError("Invalid input type")
        return async_documents

    @abstractmethod
    def lazy_transform_documents(
        self, documents: Iterator[Document], **kwargs: Any
    ) -> Iterator[Document]:
        """Transform an iterator of documents.

        Args:
            documents: A sequence of Documents to be transformed.

        Returns:
            An iterator oftransformed Documents.
        """
        raise NotImplementedError()

    @abstractmethod
    async def _alazy_transform_documents(  # type: ignore
        self, documents: AsyncIterator[Document], **kwargs: Any
    ) -> AsyncGenerator[Document, None]:
        yield None  # type: ignore

    async def alazy_transform_documents(
        self,
        documents: Input,
        **kwargs: Any,  # pylint: disable=unused-argument
    ) -> AsyncIterator[Document]:
        """Asynchronously transform an iterator of documents.

        Args:
            documents: An iterator of Documents to be transformed.

        Returns:
            An iterator of transformed Documents.
        """
        async_documents = LazyDocumentTransformer._doc_to_async_iterator(documents)

        async for doc in self._alazy_transform_documents(async_documents):
            yield doc
